@@ -7,7 +7,7 @@ All functions operate on nd arrays, where the last dimension (vectors) or
 the last two dimensions (matrices) contain individual rotations.
 """
 import numpy as np
-from .rotations import angle_between_vectors
+from .rotations import angle_between_vectors, slerp_weights
 
 
 def norm_vectors(V, out=None):
@@ -327,8 +327,8 @@ def axis_angles_from_matrices(Rs, traces=None, out=None):
     else:
         Rs_diag = Rs_diag[0]
     out[angle_close_to_pi, :3] = (
-            np.sqrt(0.5 * (Rs_diag[angle_close_to_pi] + 1.0))
-            * np.sign(out[angle_close_to_pi, :3]))
+        np.sqrt(0.5 * (Rs_diag[angle_close_to_pi] + 1.0))
+        * np.sign(out[angle_close_to_pi, :3]))
     out[angle_not_zero, :3] /= np.linalg.norm(
         out[angle_not_zero, :3], axis=-1)[..., np.newaxis]
 
@@ -517,20 +517,148 @@ def quaternion_slerp_batch(start, end, t):
 
     Returns
     -------
-    q : array-like, shape (n_steps, 4)
+    Q : array, shape (n_steps, 4)
         Interpolated unit quaternions
     """
     t = np.asarray(t)
     angle = angle_between_vectors(start, end)
-    w1, w2 = _slerp_weights(angle, t)
+    w1, w2 = slerp_weights(angle, t)
     return (w1[:, np.newaxis] * start[np.newaxis]
             + w2[:, np.newaxis] * end[np.newaxis])
 
 
-def _slerp_weights(angle, t):
-    """Weights for spherical linear interpolation (SLERP)."""
-    if angle == 0.0:
-        return np.ones_like(t), np.zeros_like(t)
-    else:
-        return (np.sin((1.0 - t) * angle) / np.sin(angle),
-                np.sin(t * angle) / np.sin(angle))
+def batch_concatenate_quaternions(Q1, Q2, out=None):
+    """Concatenate two batches of quaternions.
+
+    We use Hamilton's quaternion multiplication.
+
+    Suppose we want to apply two extrinsic rotations given by quaternions
+    q1 and q2 to a vector v. We can either apply q2 to v and then q1 to
+    the result or we can concatenate q1 and q2 and apply the result to v.
+
+    Parameters
+    ----------
+    Q1 : array-like, shape (..., 4)
+        First batch of quaternions
+
+    Q2 : array-like, shape (..., 4)
+        Second batch of quaternions
+
+    out : array, shape (..., 4), optional (default: new array)
+        Output array to which we write the result
+
+    Returns
+    -------
+    Q12 : array, shape (..., 4)
+        Batch of quaternions that represents the concatenated rotations
+
+    Raises
+    ------
+    ValueError
+        If the input dimensions are incorrect
+    """
+    Q1 = np.asarray(Q1)
+    Q2 = np.asarray(Q2)
+
+    if Q1.ndim != Q2.ndim:
+        raise ValueError("Number of dimensions must be the same. "
+                         "Got %d for Q1 and %d for Q2." % (Q1.ndim, Q2.ndim))
+    for d in range(Q1.ndim - 1):
+        if Q1.shape[d] != Q2.shape[d]:
+            raise ValueError(
+                "Size of dimension %d does not match: %d != %d"
+                % (d + 1, Q1.shape[d], Q2.shape[d]))
+    if Q1.shape[-1] != 4:
+        raise ValueError(
+            "Last dimension of first argument does not match. A quaternion "
+            "must have 4 entries, got %d" % Q1.shape[-1])
+    if Q2.shape[-1] != 4:
+        raise ValueError(
+            "Last dimension of second argument does not match. A quaternion "
+            "must have 4 entries, got %d" % Q2.shape[-1])
+
+    if out is None:
+        out = np.empty_like(Q1)
+
+    vector_inner_products = np.sum(Q1[..., 1:] * Q2[..., 1:], axis=-1)
+    out[..., 0] = Q1[..., 0] * Q2[..., 0] - vector_inner_products
+    out[..., 1:] = (Q1[..., 0, np.newaxis] * Q2[..., 1:] +
+                    Q2[..., 0, np.newaxis] * Q1[..., 1:] +
+                    np.cross(Q1[..., 1:], Q2[..., 1:]))
+    return out
+
+
+def batch_q_conj(Q):
+    """Conjugate of quaternions.
+
+    The conjugate of a unit quaternion inverts the rotation represented by
+    this unit quaternion. The conjugate of a quaternion q is often denoted
+    as q*.
+
+    Parameters
+    ----------
+    Q : array-like, shape (..., 4)
+        Unit quaternions to represent rotations: (w, x, y, z)
+
+    Returns
+    -------
+    Q_c : array, shape (..., 4,)
+        Conjugates (w, -x, -y, -z)
+    """
+    Q = np.asarray(Q)
+    out = np.empty_like(Q)
+    out[..., 0] = Q[..., 0]
+    out[..., 1:] = -Q[..., 1:]
+    return out
+
+
+def batch_quaternion_wxyz_from_xyzw(Q_xyzw, out=None):
+    """Converts from x, y, z, w to w, x, y, z convention.
+
+    Parameters
+    ----------
+    Q_xyzw : array-like, shape (..., 4)
+        Quaternions with scalar part after vector part
+
+    out : array, shape (..., 4), optional (default: new array)
+        Output array to which we write the result
+
+    Returns
+    -------
+    Q_wxyz : array-like, shape (..., 4)
+        Quaternions with scalar part before vector part
+    """
+    Q_xyzw = np.asarray(Q_xyzw)
+    if out is None:
+        out = np.empty_like(Q_xyzw)
+    out[..., 0] = Q_xyzw[..., 3]
+    out[..., 1] = Q_xyzw[..., 0]
+    out[..., 2] = Q_xyzw[..., 1]
+    out[..., 3] = Q_xyzw[..., 2]
+    return out
+
+
+def batch_quaternion_xyzw_from_wxyz(Q_wxyz, out=None):
+    """Converts from w, x, y, z to x, y, z, w convention.
+
+    Parameters
+    ----------
+    Q_wxyz : array-like, shape (..., 4)
+        Quaternions with scalar part before vector part
+
+    out : array, shape (..., 4), optional (default: new array)
+        Output array to which we write the result
+
+    Returns
+    -------
+    Q_xyzw : array-like, shape (..., 4)
+        Quaternions with scalar part after vector part
+    """
+    Q_wxyz = np.asarray(Q_wxyz)
+    if out is None:
+        out = np.empty_like(Q_wxyz)
+    out[..., 0] = Q_wxyz[..., 1]
+    out[..., 1] = Q_wxyz[..., 2]
+    out[..., 2] = Q_wxyz[..., 3]
+    out[..., 3] = Q_wxyz[..., 0]
+    return out
